
package com.asobimo.ASPurchase;

import com.android.vending.billing.IMarketBillingService;
import com.asobimo.ASPurchase.Consts.PurchaseState;
import com.asobimo.ASPurchase.Consts.ResponseCode;

import android.app.PendingIntent;
import android.app.Service;
import android.content.ComponentName;
import android.content.Context;
import android.content.Intent;
import android.content.ServiceConnection;
import android.os.Bundle;
import android.os.IBinder;
import android.os.RemoteException;
import java.security.SecureRandom;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedList;

import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;

public class BillingService extends Service implements ServiceConnection {

	private static IMarketBillingService m_Service;
	// Mҋ@Xg
	private static LinkedList<BillingRequest> m_PendingRequests = new LinkedList<BillingRequest>();
	// MɃNGXgIDmF
	private static HashMap<Long, BillingRequest> m_SentRequests = new HashMap<Long, BillingRequest>();

	private static String m_packageName="";
	public static String getAppPackageName(){return m_packageName;}

	// Avœ𐧌䂷邽߂̃nh
	private static PurchaseHandler m_purchasehandler;
	public BillingService() {
		super();
	}

	public void setContext(Context context, PurchaseHandler purchasehandler) {
		m_purchasehandler = purchasehandler;
		m_packageName = context.getPackageName();
		attachBaseContext(context);
	}

    @Override
    public IBinder onBind(Intent intent) {
        return null;
    }

    @Override
    public void onStart(Intent intent, int startId) {
        handleCommand(intent, startId);
    }

    public void handleCommand(Intent intent, int startId) {
        String action = intent.getAction();
        if ((getAppPackageName()+Consts.ACTION_CONFIRM_NOTIFICATION).equals(action)) {
            String[] notifyIds = intent.getStringArrayExtra(Consts.NOTIFICATION_ID);
            confirmNotifications(startId, notifyIds);
        } else if ((getAppPackageName()+Consts.ACTION_GET_PURCHASE_INFORMATION).equals(action)) {
            String notifyId = intent.getStringExtra(Consts.NOTIFICATION_ID);
            getPurchaseInformation(startId, new String[] { notifyId });
        } else if (Consts.ACTION_PURCHASE_STATE_CHANGED.equals(action)) {
            String signedData = intent.getStringExtra(Consts.INAPP_SIGNED_DATA);
            String signature = intent.getStringExtra(Consts.INAPP_SIGNATURE);
            purchaseStateChanged(startId, signedData, signature);
        } else if (Consts.ACTION_RESPONSE_CODE.equals(action)) {
            long requestId = intent.getLongExtra(Consts.INAPP_REQUEST_ID, -1);
            int responseCodeIndex = intent.getIntExtra(Consts.INAPP_RESPONSE_CODE,
                    ResponseCode.RESULT_ERROR.ordinal());
            ResponseCode responseCode = ResponseCode.valueOf(responseCodeIndex);
            checkResponseCode(requestId, responseCode);
        }
    }

    /**
     * MarketBillingServicẽoChs
     * @return Ȃtrue
     */
    private boolean bindToMarketBillingService() {
        try {
            Intent intent = new Intent(Consts.MARKET_BILLING_SERVICE_ACTION);
            boolean bindResult = bindService(
            		intent,
                    this,  // ServiceConnection.
                    Context.BIND_AUTO_CREATE);

            if (bindResult) {
                return true;
            } else {
                //Log.e(TAG, "Could not bind to service.");
            }
        } catch (SecurityException e) {
            //Log.e(TAG, "Security exception: " + e);
        }
        return false;
    }

    // M\bh
    /** CHECK_BILLING_SUPPORTED𑗐M
     */
    public boolean checkBillingSupported() {
        return new CheckBillingSupported().runRequest();
    }
    /** REQUEST_PURCHASE𑗐M
     * @param productId wACeID
     * @param developerPayload Cӂ̏
     * @return }[PbgɌqȂꍇfalse
     */
    public boolean requestPurchase(String productId, String productType, String developerPayload) {
        return new RequestPurchase(productId, productType, developerPayload).runRequest();
    }
    /** GET_PURCHASE_INFORMATION𑗐M
     * w̏ڍׂPURCHASE_STATE_CHANGEDƂĎ擾B
     * @param startId IN_APP_NOTIFY̒lgp
     * @param notifyIds IN_APP_NOTIFY̒lgp
     * @return false }[PbgɌqȂꍇfalse
     */
    public boolean getPurchaseInformation(int startId, String[] notifyIds) {
        return new GetPurchaseInformation(startId, notifyIds).runRequest();
    }
    /** CONFIRM_NOTIFICATIONS𑗐M
     * w擾ƂʒmB
     * PURCHASE_STATE_CHANGEDMɑԂB 𑗐M܂ŉxPURCHASE_STATE_CHANGED͑ĂB
     * @param startId PURCHASE_STATE_CHANGED̒lgp
     * @param notifyIds@PURCHASE_STATE_CHANGED̒lgp
     * @return }[PbgɌqȂꍇfalse
     */
    public boolean confirmNotifications(int startId, String[] notifyIds) {
        return new ConfirmNotifications(startId, notifyIds).runRequest();
    }
    /** RESTORE_TRANSACTIONS𑗐M
     * wԂǗACȅԂ擾B
     * VKCXg[Ȃǂ̏ꍇ̂݌ĂсA[Jŏۑ邱
     * @return }[PbgɌqȂꍇfalse
     */
    public boolean restoreTransactions() {
        return new RestoreTransactions().runRequest();
    }

    // M\bh
    /**@PURCHASE_STATE_CHANGEDMɌĂ΂
     */
	private void purchaseStateChanged(int startId, String signedData, String signature) {
		{
			if(m_purchasehandler == null){
				return;
			}
			if(!m_purchasehandler.onPurchaseStateChanged(startId, signedData, signature)){
				return;
			}
        }
        
		ArrayList<VerifiedPurchase> purchases;
        //purchases = Security.verifyPurchase(signedData, signature);
        purchases = jsonDecode(signedData);
        if (purchases == null) {
            return;
        }
        ArrayList<String> notifyList = new ArrayList<String>();
        for (VerifiedPurchase vp : purchases) {
            if (vp.notificationId != null) {
                notifyList.add(vp.notificationId);
            }
            ResponseHandler.purchaseResponse(this, vp.purchaseState, vp.productId,
                    vp.orderId, vp.purchaseTime, vp.developerPayload);
        }

        if (!notifyList.isEmpty()) {
            String[] notifyIds = notifyList.toArray(new String[notifyList.size()]);
            confirmNotifications(startId, notifyIds);
        }
    }
	
	static public ArrayList<VerifiedPurchase> jsonDecode(String signedData){
		JSONObject jObject;
        JSONArray jTransactionsArray = null;
        int numTransactions = 0;
        //long nonce = 0L;
        try {
            jObject = new JSONObject(signedData);

            // [ULZrŃAvI肵ꍇAnoncenullm܂
            //nonce = jObject.optLong("nonce");
            jTransactionsArray = jObject.optJSONArray("orders");
            if (jTransactionsArray != null) {
                numTransactions = jTransactionsArray.length();
            }
        } catch (JSONException e) {
        	//Log.e(TAG, "JSON exception1: ", e);
            return null;
        }

        //if (!Security.isNonceKnown(nonce)) {
        //    Log.w(TAG, "Nonce not found: " + nonce);
        //    return null;
        //}

        ArrayList<VerifiedPurchase> purchases = new ArrayList<VerifiedPurchase>();
        try {
            for (int i = 0; i < numTransactions; i++) {
                JSONObject jElement = jTransactionsArray.getJSONObject(i);
                int response = jElement.getInt("purchaseState");
                PurchaseState purchaseState = PurchaseState.valueOf(response);
                String productId = jElement.getString("productId");
                //String packageName = jElement.getString("packageName");
                long purchaseTime = jElement.getLong("purchaseTime");
                String orderId = jElement.optString("orderId", "");
                String notifyId = null;
                if (jElement.has("notificationId")) {
                    notifyId = jElement.getString("notificationId");
                }
                String developerPayload = jElement.optString("developerPayload", null);

                // purchase statePURCHASED̏ꍇA nonce̊mFK{
                //if (purchaseState == PurchaseState.PURCHASED) {
                //    continue;
                //}
                purchases.add(new VerifiedPurchase(purchaseState, notifyId, productId, orderId, purchaseTime, developerPayload));
            }
        } catch (JSONException e) {
            //Log.e(TAG, "JSON exception2: ", e);
            return null;
        }
        //Security.removeNonce(nonce);
        return purchases;
	}
	
	private static final SecureRandom RANDOM = new SecureRandom();
	public static long generateNonce() {
		return RANDOM.nextLong();
	}

    /**@RESPONSE_CODEMɌĂ΂
     */
    private void checkResponseCode(long requestId, ResponseCode responseCode) {
        BillingRequest request = m_SentRequests.get(requestId);
        if (request != null) {
            request.responseCodeReceived(responseCode);
        }
        m_SentRequests.remove(requestId);
    }

    /** ڑ܂Ŏs҂ԂNGXgs
     */
    private void runPendingRequests() {
        int maxStartId = -1;
        BillingRequest request;
        while ((request = m_PendingRequests.peek()) != null) {
            if (request.runIfConnected()) {
                m_PendingRequests.remove();

                if (maxStartId < request.getStartId()) {
                    maxStartId = request.getStartId();
                }
            } else {
                // ڑȂ
                bindToMarketBillingService();
                return;
            }
        }

        if (maxStartId >= 0) {
            stopSelf(maxStartId);
        }
    }

    /** MarketBillingServiceڑɌĂ΂BsUIXbh
     */
    @Override
    public void onServiceConnected(ComponentName name, IBinder service) {
        m_Service = IMarketBillingService.Stub.asInterface(service);
        runPendingRequests();
    }

    /** MarketBillingServiceؒfɌĂ΂
     */
    public void onServiceDisconnected(ComponentName name) {
        //Log.w(TAG, "Billing service disconnected");
        m_Service = null;
    }

    /** MarketBillingServicẽoCh
     * AvI܂łɌĂ΂ȂƃRlNV[NƂȂB
     */
    public void unbind() {
        try {
            unbindService(this);
        } catch (IllegalArgumentException e) {
            // This might happen if the service was disconnected
        }
    }
    
    

    /** M郊NGXg̒ۃNX
     */
    abstract class BillingRequest {
    	// Mp
        private final int m_StartId;
        // response󂯎IDB񓯊CegƂ̌ʏƍp
        protected long m_RequestId;

        public BillingRequest(int startId) {
            m_StartId = startId;
        }

        public int getStartId() {
            return m_StartId;
        }

        public boolean runRequest() {
            if (runIfConnected()) {
                return true;
            }

            if (bindToMarketBillingService()) {
                // T[rXڑꎟs
                m_PendingRequests.add(this);
                return true;
            }
            return false;
        }

        /** T[rXڑĂꍇANGXgsB
         * @return true
         */
        public boolean runIfConnected() {
            if (m_Service != null) {
                try {
                    m_RequestId = run();
                    if (m_RequestId >= 0) {
                        m_SentRequests.put(m_RequestId, this);
                    }
                    return true;
                } catch (RemoteException e) {
                    onRemoteException(e);
                }
            }
            return false;
        }

        protected void onRemoteException(RemoteException e) {
            //Log.w(TAG, "remote billing service crashed");
            m_Service = null;
        }

        abstract protected long run() throws RemoteException;

        /** RESPONSE_CODEMɌĂ΂
         */
        protected void responseCodeReceived(ResponseCode responseCode) {
        }

        protected Bundle makeRequestBundle(String method) {
            Bundle request = new Bundle();
            request.putString(Consts.BILLING_REQUEST_METHOD, method);
            request.putInt(Consts.BILLING_REQUEST_API_VERSION, 2);	// zȂFP	FQ
            request.putString(Consts.BILLING_REQUEST_PACKAGE_NAME, getPackageName());
            //Log.i(TAG, "make bundle "+method);
            return request;
        }

        protected void logResponseCode(String method, Bundle response) {
        }
    }

    /** CHECK_BILLING_SUPPORTEDNGXg̑MsNX
     */
    private class CheckBillingSupported extends BillingRequest {
        private CheckBillingSupported() {
            super(-1);
        }

        @Override
        protected long run() throws RemoteException {
            Bundle request = makeRequestBundle("CHECK_BILLING_SUPPORTED");
            request.putString(Consts.BILLING_REQUEST_ITEM_TYPE, Consts.ITEM_TYPE_SUBSCRIPTION);	// z

            Bundle response = m_Service.sendBillingRequest(request);
            int responseCode = response.getInt(Consts.BILLING_RESPONSE_RESPONSE_CODE);
            boolean billingSupported = (responseCode == ResponseCode.RESULT_OK.ordinal());
            ResponseHandler.checkBillingSupportedResponse(billingSupported);
            return Consts.BILLING_RESPONSE_INVALID_REQUEST_ID;
        }
    }

    /**@REQUEST_PURCHASENGXg̑MsNX
     */
    public class RequestPurchase extends BillingRequest {
        public final String mProductId;
        public final String mProductType;
        public final String mDeveloperPayload;

//        public RequestPurchase(String itemId) {
//            this(itemId, null);
//        }

        private RequestPurchase(String itemId, String itemType, String developerPayload) {
            super(-1);
            mProductId = itemId;
            mProductType = itemType;
            mDeveloperPayload = developerPayload;
        }

        @Override
        protected long run() throws RemoteException {
            Bundle request = makeRequestBundle("REQUEST_PURCHASE");
            request.putString(Consts.BILLING_REQUEST_ITEM_ID, mProductId);
            request.putString(Consts.BILLING_REQUEST_ITEM_TYPE, mProductType);
            if (mDeveloperPayload != null) {
                request.putString(Consts.BILLING_REQUEST_DEVELOPER_PAYLOAD, mDeveloperPayload);
            }
            Bundle response = m_Service.sendBillingRequest(request);
            PendingIntent pendingIntent
                    = response.getParcelable(Consts.BILLING_RESPONSE_PURCHASE_INTENT);
            if (pendingIntent == null) {
                //Log.e(TAG, "Error with requestPurchase");
                return Consts.BILLING_RESPONSE_INVALID_REQUEST_ID;
            }

            Intent intent = new Intent();
            ResponseHandler.buyPageIntentResponse(pendingIntent, intent);
            return response.getLong(Consts.BILLING_RESPONSE_REQUEST_ID,
                    Consts.BILLING_RESPONSE_INVALID_REQUEST_ID);
        }

        @Override
        protected void responseCodeReceived(ResponseCode responseCode) {
            ResponseHandler.responseCodeReceived(BillingService.this, this, responseCode);
        }
    }

    /**@CONFIRM_NOTIFICATIONSNGXg̑MsNX
     */
    private class ConfirmNotifications extends BillingRequest {
        final String[] mNotifyIds;

        private ConfirmNotifications(int startId, String[] notifyIds) {
            super(startId);
            mNotifyIds = notifyIds;
        }

        @Override
        protected long run() throws RemoteException {
            Bundle request = makeRequestBundle("CONFIRM_NOTIFICATIONS");
            request.putStringArray(Consts.BILLING_REQUEST_NOTIFY_IDS, mNotifyIds);
            Bundle response = m_Service.sendBillingRequest(request);
            logResponseCode("confirmNotifications", response);
            return response.getLong(Consts.BILLING_RESPONSE_REQUEST_ID,
                    Consts.BILLING_RESPONSE_INVALID_REQUEST_ID);
        }
    }

    /**@GET_PURCHASE_INFORMATIONNGXg̑MsNX
     */
    private class GetPurchaseInformation extends BillingRequest {
        long mNonce;
        final String[] mNotifyIds;

        private GetPurchaseInformation(int startId, String[] notifyIds) {
            super(startId);
            mNotifyIds = notifyIds;
        }

        @Override
        protected long run() throws RemoteException {
            mNonce = generateNonce();

            Bundle request = makeRequestBundle("GET_PURCHASE_INFORMATION");
            request.putLong(Consts.BILLING_REQUEST_NONCE, mNonce);
            request.putStringArray(Consts.BILLING_REQUEST_NOTIFY_IDS, mNotifyIds);
            Bundle response = m_Service.sendBillingRequest(request);
            logResponseCode("getPurchaseInformation", response);
            return response.getLong(Consts.BILLING_RESPONSE_REQUEST_ID,
                    Consts.BILLING_RESPONSE_INVALID_REQUEST_ID);
        }

        @Override
        protected void onRemoteException(RemoteException e) {
            super.onRemoteException(e);
            //Security.removeNonce(mNonce);
        }
    }

    /**@RESTORE_TRANSACTIONSNGXg̑MsNX
     */
    public class RestoreTransactions extends BillingRequest {
        long mNonce;

        private RestoreTransactions() {
            super(-1);
        }

        @Override
        protected long run() throws RemoteException {
            mNonce = generateNonce();

            Bundle request = makeRequestBundle("RESTORE_TRANSACTIONS");
            request.putLong(Consts.BILLING_REQUEST_NONCE, mNonce);
            Bundle response = m_Service.sendBillingRequest(request);
            logResponseCode("restoreTransactions", response);
            return response.getLong(Consts.BILLING_RESPONSE_REQUEST_ID,
                    Consts.BILLING_RESPONSE_INVALID_REQUEST_ID);
        }

        @Override
        protected void onRemoteException(RemoteException e) {
            super.onRemoteException(e);
            //Security.removeNonce(mNonce);
        }

        @Override
        protected void responseCodeReceived(ResponseCode responseCode) {
            ResponseHandler.responseCodeReceived(BillingService.this, this, responseCode);
        }
    }
    
    /** VerifywێNXB
     */
    public static class VerifiedPurchase {
        public PurchaseState purchaseState;
        public String notificationId;
        public String productId;
        public String orderId;
        public long purchaseTime;
        public String developerPayload;

        public VerifiedPurchase(PurchaseState purchaseState, String notificationId,
                String productId, String orderId, long purchaseTime, String developerPayload) {
            this.purchaseState = purchaseState;
            this.notificationId = notificationId;
            this.productId = productId;
            this.orderId = orderId;
            this.purchaseTime = purchaseTime;
            this.developerPayload = developerPayload;
        }
    }
}
